;(use bindings filepath posix)
;(load-relative "infinity.general.scm")
(cond-expand
 (compiling
  (declare (unit infinity.filesystem))
  (declare (uses filepath infinity.general)))
 (else
  (require-extension srfi-1 srfi-13 format data-structures ansi-escape-sequences basic-sequences filepath)
  (load-relative "infinity.general.scm")))

(module infinity.filesystem *
 (import chicken scheme bindings filepath posix ports files extras srfi-13 infinity.general)
 ;(import infinity.general)
 (import-for-syntax chicken scheme bindings filepath posix ports files extras srfi-13 infinity.general)
 ;(begin-for-syntax
  ;(use bindings chicken)
  ;(import infinity.general))

 (define-syntax with-open-file
  (ir-macro-transformer
   (lambda (expr inject compare)
     (bind+ (port-var filename ('key direction if-exists if-does-not-exist)) (cadr expr)
      (let ((if-exists         (or if-exists         (case direction
                                                      ((in input) #f)
                                                      (else       ''error))))
            (if-does-not-exist (or if-does-not-exist (case direction
                                                      ((in input) ''error)
                                                      (else       ''create)))))
       `(begin
         (if (file-exists? ,(inject filename))
          (case ,(inject if-exists)
           ((append))
           ((error)
            (error (string-join (list "File already exists:" ,(inject filename)))))
           ((supersede)
            (if (regular-file? ,(inject filename))
             (delete-file ,(inject filename))
             (error (string-join (list "Cannot supersede \"" ,(inject filename) "\", it is not a regular file" "")))))
           (else
            (when (eq? ,direction 'out)
             (print ,(inject if-exists))
             (error "CaseElseException"))))
          (case ,if-does-not-exist
           ((create))
           ((error)
            (error (string-join (list "File does not exist:" ,(inject filename)))))
           (else  (error "CaseElseException"))))
         (let ((,(inject port-var) (case ,direction
                                    ((in input)   (open-input-file ,(inject filename)))
                                    ((out output) (if (eq? ,(inject if-exists) 'append)
                                                   (open-output-file ,(inject filename) #:append)
                                                   (open-output-file ,(inject filename))))
                                    (else  (error "CaseElseException")))))
          ;(case ',direction
          ; ((out output)
          ;  (when (eq? ',if-exists 'append)
          ;   (set-file-position! ,(inject port-var) 0 seek/end)
          ;   (quit))))
          (let ((result (begin
                         ,@(cddr expr))))
           (flush-output ,(inject port-var))
           (close-output-port ,(inject port-var))
           result))))))))

 (define (read-file-lines filename)
  (with-open-file (port filename ('direction 'in))
   (read-lines port)))

 (define (write-file-lines filename lines #!key (if-exists 'error))
  (with-open-file (port filename ('direction 'out 'if-exists if-exists))
   (let next-line ((lines lines))
    (unless (null? lines)
     (write-line (car lines))
     (next-line (cdr lines))))
   (flush-output port)))

 (define (write-file filename content #!key (if-exists 'error))
  (with-open-file (port filename ('direction 'out 'if-exists if-exists))
   (with-output-to-port port
    (lambda ()
     (let next-form ((forms content))
      (unless (null? forms)
       (write (car forms))
       (newline)
       (next-form (cdr forms))))))
   (flush-output port)))

 (define (make-absolute-path path)
  (filepath:combine (current-directory) path))

 (define-syntax with-temporary-directory
  (ir-macro-transformer
   (lambda (expr inject compare)
    `(let* ((dir                     (create-temporary-directory))
            (,(second expr) dir))
      ,@(cddr expr)
      (delete-directory dir #t)))))

 (define-syntax with-chdir
  (ir-macro-transformer
   (lambda (expr inject compare)
    `(begin
      (let ((old-dir (current-directory)))
       (change-directory ,(cadr expr))
       (let ((result (begin ,@(cddr expr))))
        (change-directory old-dir)
        result))))))

 (define platform-exe-extension (cond
                                 ((filepath:is-windows?) ".exe")
                                 ((filepath:is-posix?)   "")
                                 (else (error "CaseElseException")))))

;(import infinity.filesystem)
;(with-open-file (port "temp.txt" ('direction out 'if-exists append))
; (print port))
