
; To see what macros are available inside a macro, evaluate:    (map car (##sys#macro-environment))
; To see what functions are available inside a macro, evaluate: (map car (##sys#meta-macro-environment))

(cond-expand
 (compiling
  (declare (unit infinity.general))
  (declare (uses matchable)))
 (else
  (require-extension srfi-1 srfi-13 format data-structures ansi-escape-sequences basic-sequences matchable)))
(import srfi-1 srfi-13 format data-structures ansi-escape-sequences basic-sequences)

;(set-read-syntax! #\{      ; transforms nested { ... } braces into (begin ...) clauses
;   (lambda (port)
;     (let loop ((c (peek-char port)) (exps '()))
;       (cond ((eof-object? c)
;              (error "EOF encountered while parsing { ... } clause"))
;             ((char=? c #\)
;              (read-char port)   ; discard
;              `(begin ,@(reverse exps)))
;             ((char-whitespace? c)
;              (read-char port)   ; discard whitespace
;              (loop (peek-char port) exps))
;             (else
;              (let ((exp (read port)))
;                (loop (peek-char port)
;                      (cons exp exps))))))))

(module infinity.general *
 (import scheme chicken srfi-1 srfi-13 data-structures ansi-escape-sequences basic-sequences bindings ports format)
 ;(use (rename srfi-1 (take _take) (drop _drop)))
 (begin-for-syntax
  (import chicken basic-sequences bindings)
  (define-syntax with-gensyms
   (er-macro-transformer
    (lambda (expr inject compare)
     (let ((names (cadr expr))
           (body  (cddr expr)))
      `(let ,(map (lambda (n)
                   `(,n (gensym (string-upcase (->string ',n)))))
                  names)
        ,@body))))))

 (define ansi-codes '((|N| . "[00m")   (|W| . "[1;37m") (|K| . "[0;37m") (|k| . "[1;30m")
                      (|r| . "[0;31m") (|R| . "[1;31m") (|g| . "[0;32m") (|G| . "[1;32m") (|b| . "[0;34m") (|B| . "[1;34m")
                      (|c| . "[0;36m") (|C| . "[1;36m") (|m| . "[0;35m") (|M| . "[1;35m") (|y| . "[0;33m") (|Y| . "[1;33m") (|Z| . "[0;30m")
                      (|*| . "[1m")    (|_| . "[4m")
                      (|<| . "[1D")    (|>| . "[1C") (|^| . "[1A") (|v| . "[1B")))
 (define *formatting-delimiters* (make-parameter '("⟨" . "⟩")))

 (define-syntax pop-chars!
  (syntax-rules ()
   ((_ string pop-char-count)
    (let ((prelude (substring string 0 pop-char-count)))
     (set! string (substring string pop-char-count (string-length string)))
     prelude))))

 (define-syntax extend
  (er-macro-transformer
   (lambda (expr rename compare)
    `(append ,(second expr) (list ,(third expr))))))
 ;(define-syntax extend-to
 ; (er-macro-transformer
 ;  (lambda (expr rename compare)
 ;   `(set! ,(second expr) (extend ,(second expr) ,(third expr))))))
 ;(define-syntax append-to
 ; (er-macro-transformer
 ;  (lambda (expr rename compare)
 ;   `(set! ,(second expr) (append ,(second expr) ,(third expr))))))
 (define-syntax extend-to
  (syntax-rules ()
   ((_ lst value)
    (set! lst (extend lst value)))))
 (define-syntax append-to
  (syntax-rules ()
   ((_ lst values)
    (set! lst (append lst values)))))

 (define-syntax symbol-drop-right
  (syntax-rules ()
   ((_ symbol ncount)
    (string->symbol (string-drop-right (->string symbol) ncount)))))

 (define-syntax inc!
  (syntax-rules ()
    ((_ var)
     (set! var (+ var 1)))
    ((_ var amount)
     (set! var (+ var amount)))))
 (define-syntax dec!
  (syntax-rules ()
    ((_ var)
     (set! var (- var 1)))
    ((_ var amount)
     (set! var (- var amount)))))

 (define-syntax take*         ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst count)
     (take lst (min count (length lst))))))
 (define-syntax drop*         ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst count)
     (drop lst (min count (length lst))))))
 (define-syntax take-right*   ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst count)
     (take-right lst (min count (length lst))))))
 (define-syntax drop-right*   ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst count)
     (drop-right lst (min count (length lst))))))

 (define-syntax car*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (and (not (null? lst))
          (list? lst)
      (car lst)))))
 (define-syntax cdr*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (if (and (not (null? lst))
              (list? lst))
      (cdr lst)
      '()))))
 (define-syntax cadr*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (car* (cdr* lst)))))
 (define-syntax cdar*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (cdr* (car* lst)))))
 (define-syntax cddr*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (cdr* (cdr* lst)))))
 (define-syntax caadr*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (car* (cadr* lst)))))
 (define-syntax caddr*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (car* (cddr* st)))))
 (define-syntax cadar*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (car* (cdar* lst)))))
 (define-syntax cddar*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (cdr* (cdar* lst)))))
 (define-syntax caaar*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (car* (caar* lst)))))
 (define-syntax cdddr*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (cdr* (cddr* lst)))))
 (define-syntax cadddr*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (car* (cdddr* lst)))))
 (define-syntax caaddr*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (car* (caddr* lst)))))
 (define-syntax caaadr*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (car* (caadr* lst)))))
 (define-syntax caaaar*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (car* (caaar* lst)))))
 (define-syntax cdaddr*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (cdr* (caddr* lst)))))
 (define-syntax cdaadr*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (cdr* (caadr* lst)))))
 (define-syntax cdaaar*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (cdr* (caaar* lst)))))
 (define-syntax cddadr*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (cdr* (cdadr* lst)))))
 (define-syntax cddaar*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (cdr* (cdaar* lst)))))
 (define-syntax cdddar*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (cdr* (cddar* lst)))))
 (define-syntax caddar*            ; don't crash when there aren't enough elements
  (syntax-rules ()
    ((_ lst)
     (car* (cddar* lst)))))

 ;(define-syntax take-where
 ; (syntax-rules ()
 ;  ((_ predicate lst)
 ;   (remove lst (lambda (item) (not (predicate item)))))))
 (define (take-where predicate lst)
  (remove (lambda (item) (not (predicate item))) lst))

 (define-syntax ends-with?
   (syntax-rules ()
     ((_ lst what)
      (equal? (take-right* lst (length what)) what))))

 (define-syntax join-strings
  (ir-macro-transformer
   (lambda (expr inject compare)
    `(string-join (list ,@(cdr expr)) ""))))

 ;(define-syntax pop-chars
 ; (ir-macro-transformer
 ;  (lambda (expr inject compare)
 ;   ;(with-gensyms (string count))
 ;   (bind (string count) (cdr expr)
 ;    ;(print string)
 ;    ;(print count)
 ;    `(let* ((string-value ,string)
 ;            (popped (substring/shared string-value 0 ,count)))
 ;      (print string-value)
 ;      (print popped)
 ;      (print ,(inject string)))))))
 ;    ;  (set! ,(inject string) (substring/shared string-value count))
 ;    ;  popped)))))

 (define-syntax bind+    ;; NOTE: Do not mix 'key and 'opt in the same scope! It will not work!
  (ir-macro-transformer
   (lambda (expr inject compare)
    (with-gensyms (resume-data)
     (if (null? (cadr expr))                                 ; if no pattern remains
      `(begin
        ;(newline) (print "*1")
        (print ",(caddr expr) = " ,(caddr expr))
        (if (not (null? ,(caddr expr)))
         (error "Ran out of pattern (1)"))                   ; if no pattern remains, but some data remains, fail
        ,@(cdddr expr))                                      ; if no pattern remains, and no data remains, just execute the body
      (let ((current-pattern (caadr expr)))
       ;(print expr)
       ;(print current-pattern " =~ " ''key ": " (compare current-pattern ''key))
       (cond
        ((compare current-pattern ''key)
         (begin
          ;(newline) (print "*key: " expr)
          (if (null? (cdadr expr))
           `(error "Ran out of pattern (2)")
           (let* ((current-pattern            (cadadr expr))
                  (wrapped-pattern            (list current-pattern))
                  (remaining-patterns         (cddadr expr))
                  (wrapped-remaining-patterns (append '('key) (cddadr expr))))
            `(let* ((evaluated-data ,(caddr expr))
                    (data-tail      (find-tail (lambda (d)
                                                (,compare d '',current-pattern))
                                               evaluated-data))
                    (data           (if data-tail (cadr data-tail) #f))
                    (,resume-data   evaluated-data))
              ;(print ',current-pattern " = " data)
              ;(print "wrapped-pattern: " ',wrapped-pattern)
              ;(print "(bind+ " ',wrapped-pattern " " (list data) ")")
              (bind+ ,wrapped-pattern (list data)
               ;(print "*5 " ',current-pattern)
               ;(print ',remaining-patterns)
               (if (null? ',remaining-patterns)               ; if no pattern remains, just execute the body
                (begin
                 ,@(cdddr expr))
                (begin
                 ;(print "remaining-patterns: " ',remaining-patterns)
                 ;(print "resume-data = " ,resume-data)
                 ;(print "(bind+ " ',wrapped-remaining-patterns " " ,resume-data ")")
                 (bind+ ,wrapped-remaining-patterns ,resume-data   ; resume binding current level, as keyword parameters
                  ,@(cdddr expr))))))))))
        ((compare current-pattern ''opt)
         (if (null? (cdadr expr))
          `(error "Ran out of pattern (3)")
          (let* ((current-pattern            (cadadr expr))
                 (wrapped-pattern            (list current-pattern))
                 (remaining-patterns         (cddadr expr))
                 (wrapped-remaining-patterns (append '('opt) (cddadr expr))))
            `(let* ((evaluated-data ,(caddr expr))
                    (,resume-data   (if (null? evaluated-data) '() (cdr evaluated-data)))
                    (data           (if (null? evaluated-data) '() (car evaluated-data))))
              ;(newline) (print "*opt: " ',expr)
              ;(print "wrapped-pattern: " ',wrapped-pattern)
              ;(print "evaluated-data: " evaluated-data)
              ;(print "resume-data: " ,resume-data)
              ;(print "data: " data)
              ;(print "(bind+ " ',wrapped-pattern " " (list data) ")")
              (bind+ ,wrapped-pattern (list data)
               ;(print "*6 " ',current-pattern)
               ;(print ',remaining-patterns)
               (when (not (null? ',remaining-patterns))               ; if no pattern remains, just execute the body
                ;(print "(bind+ " ',wrapped-remaining-patterns " " ,resume-data ")")
                (bind+ ,wrapped-remaining-patterns ,resume-data
                 ,@(cdddr expr)))
               ,@(cdddr expr))))))
        ((list? current-pattern)                             ; if current pattern is a subpattern, recurse
         `(begin
           ;(newline) (print "*2: " ',expr)
           ;(print "expr = " ',expr)
           (let* ((evaluated-data ,(caddr expr))
                  (sub-data       (car evaluated-data))
                  (,resume-data   (cdr evaluated-data)))
            ;(print "evaluated-data = " evaluated-data)
            ;(print "sub-data = " sub-data)
            ;(print "resume-data = " ,resume-data)
            ;(print "(bind+ " ',current-pattern " " sub-data ")")
            (bind+ ,current-pattern sub-data      ; bind nested pattern
             ;(print "*4 " ',current-pattern)
             (if (and (null? ',(cdadr expr))                  ; if no pattern remains, and no data remains, just execute the body
                      (null? ,resume-data))
              (begin
               ,@(cdddr expr))
              (begin
               ;(print "resume-data = " ,resume-data)
               ;(print "(bind+ " ',(cdadr expr) " " ,resume-data ")")
               (bind+ ,(cdadr expr) ,resume-data               ; otherwise, resume binding current level
                ,@(cdddr expr))))))))
        (else
         `(begin
           ;(newline) (print "*3")
           ;(print "expr = " ',expr)
           (let ((evaluated-data ,(caddr expr)))
            (if (null? evaluated-data)                        ; we should have data here
             (error "Ran out of data (1)")
             (begin
              ;(print "expr = " ',expr)
              ;(print "data = " ',(cddr expr))
              ;(print "evaluated-data = " evaluated-data)
              (let ((,current-pattern (,inject (car evaluated-data))))   ; create a binding for this pattern
               ;(print ',current-pattern " = " ,current-pattern)
               (cond
                ((null? ',(cdadr expr))
                 (if (not (or (null? evaluated-data)
                              (null? (cdr evaluated-data))))
                  (error "Ran out of pattern (4)"))            ; if no pattern remains, but some data remains, fail
                 ,@(cdddr expr))                               ; if no pattern remains, and no data remains, just execute the body
                ((null? evaluated-data)
                 (error "Ran out of data (2)"))                ; if no data remains, but some pattern remains, fail
                (else
                 (if (not (null? ',(cdadr expr)))                ; if further patterns exist, process them recursively
                  ;(print "(bind+ " ',(cdadr expr) " " (cdr evaluated-data) ")")
                  (bind+ ,(cdadr expr) (cdr evaluated-data)
                   ,@(cdddr expr))))))))))))))))))

 (define (split-string-to-lines string)
  (string-split string (string #\newline)))

 (define (repeat-string count string)
  (string-join (make-list count string) ""))

 (define (format-ansi port formatstring . args)
  (unless (or (port? port)
              (eq? port #t)
              (eq? port #f))
   (error "First argument must be a port, #t or #f"))
  (define (translate-ansi-code code)
   (string-join (list (string #\esc) (alist-ref (string->symbol (->string code)) ansi-codes)) ""))
  (define (assemble-ansi-string text #!optional (closing-code (translate-ansi-code 'N)))
   (with-output-to-string
    (lambda ()
     (unless (string-null? text)
      (cond
       ((string-prefix? (car (*formatting-delimiters*)) text)
        (pop-chars! text (string-length (car (*formatting-delimiters*))))
        (display (translate-ansi-code (pop-chars! text 1))))
       ((string-prefix? (cdr (*formatting-delimiters*)) text)
        (pop-chars! text (string-length (cdr (*formatting-delimiters*))))
        (display closing-code))
       (else
        (display (pop-chars! text 1))))
      (display (assemble-ansi-string text closing-code))))))
  (let ((result (format port "~a"
                 (assemble-ansi-string (apply format #f formatstring args)))))
   (if port
    (values)
    result)))

 (define (true? value)
  (not (not value))))

;(import infinity.general)
;(format-ansi #t "⟨Gwoo!⟩ ⟨YSqueak!⟩~%")

;(let ((data '('a 2)))
; (bind+ ('key a b) data
;  (print a b "woo")))
;(let ((data '()))
; (bind+ ('opt a b) data
;  (print a " " b " woo")))
;(let ((data '(1 ((2 3)) 4 5)))
; (bind+ (a ((b c)) d e) data
;  (print a b c d e "woo")))
