(require-extension posix srfi-1 srfi-13 srfi-19-date srfi-19-io format loops data-structures files)
(load-relative "infinity.general.scm")
(load-relative "infinity.filesystem.scm")

(cond-expand
 (compiling)
 (else))
  ;(use debug)))
  ;(trace member)))

(module infinity.logging (log/add-file-output log/add-console-output log/console+ log/debug+ log/warn+ log/info+ log/error+)
 (import scheme chicken srfi-1 srfi-13 srfi-19 data-structures format posix filepath files infinity.general infinity.filesystem)
 (use loops)

 (define *log-outputs* (list))
 (define *log-indent*  0)

 (define (log/add-console-output levels-list)
  (extend-to *log-outputs* (list levels-list 'console)))

 (define (log/add-file-output levels-list #!key (logfile "logfile.log") (if-exists 'rename) (max-logfiles 10))
  (extend-to *log-outputs* (list levels-list 'file (make-absolute-path logfile)))
  (cond
   ((regular-file? logfile)
    (case if-exists
     ((clobber)
      (delete-file logfile))
     ((rename)
      (let ((base         (filepath:drop-extension logfile))
            (ext          (filepath:take-extension logfile))
            (file-time    (file-modification-time logfile)))
       (when max-logfiles
        (let* ((date-part    (format-date "~Y~m~d#~H~M~S" (seconds->date file-time)))
               (new-filename (string-join (list base "-" date-part ext) ""))
               (count        0))
         (do-while (file-exists? new-filename)
          (inc! count)
          (set! new-filename (string-join (list base "-" date-part "." (->string count) ext) "")))
         (file-move logfile new-filename))
        (let* ((old-logs     (sort (glob (string-join (list base "-*" ext) ""))
                                   (lambda (f1 f2)
                                    (< (file-modification-time f1) (file-modification-time f2)))))
               (expired-logs (drop-right* old-logs (- max-logfiles 1))))
         ;(print old-logs)
         ;(print expired-logs)
         (do-list expired-file expired-logs
          (delete-file expired-file))))))))
   ((file-exists? logfile)
    (error (string-join "Cannot log to \"" logfile "\": a non-regular file is in the way.")))))

 (define (log/debug+ . message)
  (log-internal 'debug message))
 (define (log/info+ . message)
  (log-internal 'info message))
 (define (log/warn+ . message)
  (log-internal 'warn message))
 (define (log/error+ . message)
  (log-internal 'error message))
 (define (log/console+ . message)
  (format #t "~a" (format-log-message message)))

 (define (format-log-message message)
  (if (string? (car message))
   (set! message (apply format-ansi #f message)))
  (set! message (string-intersperse (map (lambda (line)
                                          (conc (repeat-string *log-indent* #\space) line))
                                         (string-split message (string #\newline)))
                                    (string #\newline)))
  (if (string-suffix? (string #\newline) message)
   message
   (conc message #\newline)))

 (define (log-internal level message-list)
  (when (null? *log-outputs*)
   (error "Call (add-file-output) or (add-console-output) first!"))
  (let ((message (format-log-message message-list)))
   (unless (string-null? message)
    (let do-next-output ((outputs (take-where (lambda (item)
                                               (member level (car item) equal?))
                                              *log-outputs*)))
     (unless (null? outputs)
      (let ((output-type   (cadar* outputs))
            (output-target (caddar* outputs)))
       (case output-type
        ((console)
         (format #t "~a" message))
        ((file)
         (with-open-file (port output-target ('direction 'output 'if-exists 'append 'if-does-not-exist 'create))
          (format port "~a" message)))
        (else (error "CaseElseException"))))
      (do-next-output (cdr outputs))))))))

;(import infinity.logging)
;(init-log logfile: "test.log" max-logfiles: 1)
;(log+ "⟨r~a⟩" 'woo)

 ;(define (init-log #!key (logfile "logfile.log") (log-to-console #f) (if-exists 'rename) (max-logfiles 10))
 ; (when log-to-console
 ;  (extend-to *log-outputs* #t))
 ; (when logfile
 ;  (extend-to *log-outputs* (make-absolute-path logfile))
 ;  (cond
 ;   ((regular-file? logfile)
 ;    (case if-exists
 ;     ((clobber)
 ;      (delete-file logfile))
 ;     ((rename)
 ;      (let ((base         (filepath:drop-extension logfile))
 ;            (ext          (filepath:take-extension logfile))
 ;            (file-time    (file-modification-time logfile)))
 ;       (when max-logfiles
 ;        (let* ((date-part    (format-date "~Y~m~d#~H~M~S" (seconds->date file-time)))
 ;               (new-filename (string-join (list base "-" date-part ext) ""))
 ;               (count        0))
 ;         (do-while (file-exists? new-filename)
 ;          (inc! count)
 ;          (set! new-filename (string-join (list base "-" date-part "." (->string count) ext) "")))
 ;         (file-move logfile new-filename))
 ;        (let* ((old-logs     (sort (glob (string-join (list base "-*" ext) ""))
 ;                                   (lambda (f1 f2)
 ;                                    (< (file-modification-time f1) (file-modification-time f2)))))
 ;               (expired-logs (drop-right* old-logs (- max-logfiles 1))))
 ;         ;(print old-logs)
 ;         ;(print expired-logs)
 ;         (do-list expired-file expired-logs
 ;          (delete-file expired-file))))))))
 ;   ((file-exists? logfile)
 ;    (error (string-join "Cannot log to \"" logfile "\": a non-regular file is in the way."))))))
