#!/usr/bin/csi -script

(use srfi-1 srfi-13 extras shell filepath loops environments)
;(use http-client)
(cond-expand
 (compiling)
  ;(declare (uses infinity.general infinity.filesystem infinity.logging)))
 (else
  (load-relative "include.scm/infinity.general.scm")
  (load-relative "include.scm/infinity.filesystem.scm")
  (load-relative "include.scm/infinity.logging.scm")))
  ;(use debug)))
(import infinity.general infinity.filesystem infinity.logging)

(define got-packages (list))
;(define import-libs (list))

(define (compile filename)
 ;(shell-verbose #t)
 (let* ((filename    (make-absolute-path filename))
        (output-dir  (filepath:take-directory filename)))
        ;(output-file (pathname-replace-extension (filepath:combine output-dir (filepath:take-file-name filename)) platform-exe-extension)))
  (with-temporary-directory tempdir
    (log/debug+ tempdir)
    (with-chdir tempdir
     ;(let ((filename (sanitise-and-copy-file filename)))
     ;(let ((filename (filepath:take-file-name filename)))
     (compile-file filename output-dir force?: #t static?: #t))))
 (format-ansi #t "⟨GDone⟩~%"))

(define (compile-file filename output-dir #!key force? static?)
 ;(let* ((output-dir  (filepath:take-directory filename))
        ;(output-file (pathname-replace-extension
        ;              (filepath:combine output-dir (filepath:take-file-name filename))
        ;              platform-exe-extension))
 (let* ((unit-declaration (find-unit-declaration filename (read-file filename)))
        (make-exe?        (not unit-declaration))
        (output-file      (filepath:combine output-dir
                                            (filepath:replace-extension (filepath:take-file-name filename)
                                                                        (if make-exe? platform-exe-extension "o")))))
  ;(print "UNIT: " unit-declaration)
  (if (and (not make-exe?)
           (not force?)
           (regular-file? output-file))
   (begin
    (log/info+ "⟨k~a has already been compiled⟩~%" filename)
    '())
   (begin
    (log/debug+   "Preparing to compile ⟨B~a⟩ to ⟨W~a⟩~%" filename output-file)
    (log/console+ "Preparing to compile ⟨B~a⟩~%" (filepath:take-file-name filename))
    (let* ((object-files   (delete '() (flatten (prepare-dependancies-for filename)))))
     ;(print object-files)
     (let ((cmdline        (string-join (flatten (delete #f (list "csc -d2 -O0"
                                                                  (and static? "-static")
                                                                  (if make-exe?
                                                                   (string-join (map ->string got-packages) "-uses " 'prefix)
                                                                   (list "-c -j" (filepath:take-base-name filename)))
                                                                  filename object-files "-o" output-file))))))
      ;(print "Created object files: " (string-join object-files))
      (log/debug+   " Compiling ⟨B~a⟩ to ⟨W~a⟩~%" filename output-file)
      (log/console+ " Compiling ⟨B~a⟩ to ⟨W~a⟩~%" (filepath:take-file-name filename) (filepath:take-file-name output-file))
      (log/debug+ "⟨k~a⟩~%" cmdline)
      (exec cmdline)
      output-file))))))

;(define (sanitise-and-copy-file filename)
; (let* ((source-dir  (filepath:take-directory filename))
;        (source-code (map (lambda (form)
;                           (translate-toplevel-form source-dir form))
;                          (read-file filename)))
;        (output-file (filepath:add-extension (join-strings (filepath:take-base-name filename) "%") "scm")))
;  (log+ "Sanitising ⟨B~a⟩ into ⟨B~a⟩~%" filename output-file)
;  (write-file output-file source-code 'if-exists 'supersede)
;  output-file))

(define (translate-toplevel-form source-dir form)
 (case (and (not (null? form))
            (first form))
  ((load)
   (error "Why the hell are you using 'load'?"))
  ((load-relative)
   (read-file (filepath:combine source-dir (second form))))
  (else form)))

(define (find-depends-forms meta-content)
 (let ((depends (assoc 'depends (car meta-content))))
  (if depends
   (map (lambda (d)
         (if (list? d)
          (car d)
          d))
        (cdr depends))
   '())))

(define (prepare-dependancies-for source-file #!optional meta-file)
 ;(print "Processing " source-file)
 ;(print "(meta-file = " meta-file ")")
 (if (and meta-file
          (regular-file? meta-file))
  (begin
   ;(print " Looking for imports in " meta-file)
   (let* ((packages (find-depends-forms (read-file meta-file)))
          (object-files (delete '() (flatten (map prepare-package packages)))))
    object-files))
  (begin
   ;(print " Looking for imports in " source-file)
   (let* ((forms    (read-file source-file))
          ;(packages (append (find-package-forms-recursively source-file forms) (find-uses-declarations source-file forms))))
          (packages (append (find-package-forms-recursively source-file forms))))
    (log/debug+ " Dependencies: " (string-join (map ->string packages)))
    (let ((object-files (delete '() (flatten (map (lambda (dep)
                                                   (cond
                                                    ((string? dep)
                                                     ;(print dep " is a string")
                                                     ;(extend-to import-libs (filepath:take-base-name dep))
                                                     (compile-file dep (current-directory)))
                                                    ((symbol? dep)
                                                     ;(print dep " is a symbol")
                                                     ;(extend-to import-libs (->string dep))
                                                     (prepare-package dep))
                                                    (else '())))
                                                  packages)))))
     ;(unless (every string? object-files)
     ; (print "FOUND SOME NON-SYMBOLS (2)")
     ; (print object-files)
     ; (exit))
     object-files)))))

(define (find-unit-declaration source-file forms)
 ;(newline) (print forms)
 (let ((keywords (list 'cond-expand 'compiling '(not compiling))))
  (cond
   ((or (null? forms)
        (atom? forms))
    #f)
   ((and (atom?  (car forms))
         (member (car forms) keywords)
         (> (length forms) 1))
    (find-unit-declaration source-file (cdr forms)))
   ((and (eq?   (car forms)    'declare)
         (>     (length forms) 1)
         (list? (cadr forms))
         (eq?   (caadr forms)  'unit))
    (cadadr forms))
   (list? (car forms)
    (or (find-unit-declaration source-file (car forms))
        (find-unit-declaration source-file (cdr forms))))
   (else #f))))

;(define (find-uses-declarations source-file forms)
; ;(newline) (print forms)
; (let ((keywords (list 'cond-expand 'compiling '(not compiling))))
;  (cond
;   ((or (null? forms)
;        (atom? forms))
;    '())
;   ((and (atom?  (car forms))
;         (member (car forms) keywords)
;         (> (length forms) 1))
;    (find-uses-declarations source-file (cdr forms)))
;   ((and (eq?   (car forms)    'declare)
;         (>     (length forms) 1)
;         (list? (cadr forms))
;         (eq?   (caadr forms)  'uses))
;    (cdadr forms))
;   (list? (car forms)
;    (append (find-uses-declarations source-file (car forms))
;            (find-uses-declarations source-file (cdr forms))))
;   (else '()))))

(define (find-package-forms-recursively source-file forms)
 ;(newline) (print source-file) (print forms) (print (length forms))
 (let ((keywords (list 'use 'require-extension 'cond-expand)))
  (cond
   ((null? forms)
    (list))
   ((and (atom?  (first forms))
         (member (first forms) keywords)
         (> (length forms) 1))
    ;(print "1: " forms)
    (if (and (eq?     (car forms)             'cond-expand)
             (or      (equal?  (caadr forms)  '(not compiling))
                      (equal?  (caadr forms)  'compiling)))
     (let next-form ((forms (cdr forms)))
      ;(print "2: " forms)
      (cond
       ((or (null? forms)
            (atom? forms))
        '())
       ((or (eq? (car forms) 'load-relative)
            (eq? (car forms) 'load))
        (list (filepath:combine (filepath:take-directory source-file) (cadr forms))))
       ((list? (car forms))
        (append (next-form (car forms))
                (next-form (cdr forms))))  ; recurse
       (else
        (append (find-package-forms-recursively source-file (cdr forms))
                (next-form (cdr forms))))))  ; recurse
     ;((load-relative load)
     ; (list (make-absolute-path (filepath:combine (filepath:take-directory source-file) (second forms))))
     (cdr forms)))
   ((list? (car forms))
    (append (find-package-forms-recursively source-file (car forms))
            (find-package-forms-recursively source-file (cdr forms))))
   (else '()))))
    ;(find-package-forms-recursively source-file (cdr forms))))))

(define (exec . command)
 (let ((cmdline (string-join command)))
  (let ((exit-status  (execute (list cmdline) status: #t)))
   (unless (= 0 exit-status)
    (error (format-ansi #f "Command ⟨W~a⟩ returned an exit status of ⟨R~a⟩~%" cmdline (->string exit-status)))))))

(define (generate-compile-opts package-name)
 (case (string->symbol package-name)
  ((lookup-table)
   "-prelude \"(define-constant MAGIC-LIMIT 12)\" ")
  (else "")))

;(define (read-compile-opts package-name setup-file)
; (if (regular-file? setup-file)
;  (begin
;   (print "Loading setup file: " setup-file)
;   (let* ((file-content (read-file setup-file))
;          (compile-opts (list))
;          (program      `(begin
;                          (use versions)
;                          (define-syntax compile
;                            (syntax-rules ()
;                              ((_ . args)
;                               (begin))))
;                          (define (verify-extension-name . args) #f)
;                          (define (install-extension . args) #f)
;                          (define (standard-extension . args) #f)
;                          (define (setup-shared-extension-module . args)
;                           (when (string= (car args))
;                            (let ((opts (member compile-options: (cdr args))))
;                             (when opts
;                              (set! compile-opts (cadr opts))))))
;                          . ,file-content)))
;     (pp program)
;     (eval program)
;     (unless (null? compile-opts)
;      (print "Found compiler options: " compile-opts))
;     ;(exit)
;     compile-opts))
;  '()))

(define (get-scm-files-from-meta-file package-name meta-file)
 (let* ((meta-content (read-file meta-file))
        (files        (take-where
                       (lambda (file) (and (string-ci= (filepath:take-extension file) ".scm")
                                           (regular-file? (filepath:combine package-name file))))
                       (cdr (assoc 'files (car meta-content))))))
  ;(print "meta-file mentioned existing files: " files)
  files))

(define (prepare-package package)
 (case package
  ((chicken)
   ;(print "Not getting package " (->string package))
   '())
  (else
   (cond
    ((member package got-packages)
     ;(print "Already got package " (->string package)))
     '())
    (else
     (extend-to got-packages package)
     (log/debug+   " Getting package ⟨M~a⟩~%" (->string package))
     (log/console+ " Getting and compiling package ⟨M~a⟩~%" (->string package))
     (let* ((package-name  (->string package))
            (main-scm-file (join-strings package-name ".scm"))
            ;(object-file   (join-strings package-name ".o"))
            (meta-file     (make-absolute-path (filepath:combine package-name (join-strings package-name ".meta"))))
            (setup-file    (make-absolute-path (filepath:combine package-name (join-strings package-name ".setup")))))
      (exec "chicken-install >/dev/null -retrieve" package-name)
      (if (directory? package-name)
       ;(let* ((scm-files    (if (regular-file? main-scm-file)
       ;                      (list main-scm-file)
       ;                      (map make-absolute-path (glob (filepath:combine package-name "*.scm")))))
       ;       (object-files (flatten (map (lambda (f) (prepare-dependancies-for (make-absolute-path f) meta-file)) scm-files))))
       ; (cond
       ;  ((regular-file? object-file)
       ;   (print " Package " package-name " has already been compiled"))
       ;  (else
       ;   (print (join-strings " Compiling " package-name " into " object-file))
       ;   (with-chdir package-name
       ;    (compile-static package-name)))
       ;  (append object-files (list object-file)))))))

       (let ((scm-files (if (regular-file? (filepath:combine package-name main-scm-file))
                         (list main-scm-file)
                         (get-scm-files-from-meta-file package-name meta-file))))
                         ;(with-chdir package-name (glob "*.scm")))))
        ;(display "scm-files: ") (write scm-files) (newline)
        ;(print "scm-files: " (string-join scm-files))
        (if (null? scm-files)
         (error (join-strings " Could not find main package file(s) for \"" package-name "\". Only found "
                              (->string (directory package-name))))
         (let process-next-scm-file ((scm-files    scm-files)
                                     (object-files (list)))
          (let* ((scm-file          (car scm-files))
                 (compile-opts      (generate-compile-opts package-name))
                 ;(compile-opts  (string-join (read-compile-opts package-name setup-file) " " 'suffix))
                 (object-files      (append object-files (prepare-dependancies-for (make-absolute-path scm-file) meta-file)))
                 (object-file       (if (> (length scm-files) 1)
                                     (join-strings package-name "+" (filepath:replace-extension scm-file ".o"))
                                     (join-strings package-name ".o")))
                 (import-file-name  (join-strings package-name ".import.scm"))
                 (import-file       (filepath:combine package-name import-file-name))
                 (cmdline           (join-strings "csc -d2 -O0 " compile-opts "-unit " package-name " -j " package-name
                                              " -c " scm-file " -o ../" object-file " -I."))
                 (results           (cond
                                     ((regular-file? object-file)
                                      ;(print " File " scm-file " has already been compiled")
                                      object-files)
                                     (else
                                      (log/debug+ " Compiling ⟨B~a⟩ into ⟨W~a⟩~%" scm-file object-file)
                                      (with-chdir package-name
                                       ;(compile-static ""))
                                       (log/debug+ "⟨k~a⟩~%" cmdline)
                                       (exec cmdline))
                                      (if (regular-file? import-file)
                                       (file-move import-file import-file-name))
                                      (append object-files (list object-file))))))
           (if (null? (cdr scm-files))
            results
            (process-next-scm-file (cdr scm-files) results))))))
       '())))))))

;(cond-expand)
 ;((not compiling)))
  ;(trace find-uses-declarations)))
;  (trace find-package-forms-recursively)
;  (trace eq?)))
  ;(trace find-unit-declaration)
  ;(trace eq?)))

;(init-log log-to-console: #t)
(log/add-file-output    '(debug info warn error) "static-compile.log")
(log/add-console-output '(info warn error))
(if (>= (length (argv)) 4)
 (compile (fourth (argv))))
